#include "VirtualUSBFlashDrive.h"

NTSTATUS DriverEntry (IN OUT PDRIVER_OBJECT DriverObject, IN PUNICODE_STRING RegistryPath);
NTSTATUS VDevCreateClose (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp);
NTSTATUS VDevControl (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp);
NTSTATUS VDevReadWrite (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp);
NTSTATUS FileDiskCreateDevice (IN PDRIVER_OBJECT   DriverObject);
VOID VDevUnload (IN PDRIVER_OBJECT DriverObject);
NTSTATUS VDevFormatDiskToFat(IN PDEVICE_OBJECT DeviceObject);

NTSTATUS DriverEntry (IN OUT PDRIVER_OBJECT DriverObject, IN PUNICODE_STRING RegistryPath){
	NTSTATUS 				Status;
	UNICODE_STRING 			paramPath;
	PVDEV_DRIVER_EXTENSION	driverExt;
	
	Status = IoAllocateDriverObjectExtension(DriverObject,VDEV_DRIVER_EXTENSION_KEY,sizeof(VDEV_DRIVER_EXTENSION),&driverExt);
										 
	DriverObject->MajorFunction[IRP_MJ_CREATE]			=VDevCreateClose;
	DriverObject->MajorFunction[IRP_MJ_CLOSE]			=VDevCreateClose;
	DriverObject->MajorFunction[IRP_MJ_READ]			=VDevReadWrite;
	DriverObject->MajorFunction[IRP_MJ_WRITE]			=VDevReadWrite;
	DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL]	=VDevControl;
	DriverObject->DriverUnload 							=VDevUnload;
	
	FileDiskCreateDevice(DriverObject);
	
	return Status;
}
	
NTSTATUS VDevCreateClose (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp){
	NTSTATUS	Status;
	Status 		= STATUS_SUCCESS;
	
	Irp->IoStatus.Status = STATUS_SUCCESS;
	Irp->IoStatus.Information = 0;
	IoCompleteRequest (Irp, IO_NO_INCREMENT);
		
	return STATUS_SUCCESS;
}
	
NTSTATUS VDevControl (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp){
	PVDEV_EXTENSION   devExt = DeviceObject->DeviceExtension;
	PIO_STACK_LOCATION irpSp;
	NTSTATUS Status;
	
	Status = STATUS_IO_DEVICE_ERROR;
	
	irpSp = IoGetCurrentIrpStackLocation(Irp);
	
	switch (irpSp->Parameters.DeviceIoControl.IoControlCode)
	{

		case IOCTL_DISK_GET_MEDIA_TYPES:
		case IOCTL_DISK_GET_DRIVE_GEOMETRY:
			{
				PDISK_GEOMETRY outputBuffer;
				outputBuffer = (PDISK_GEOMETRY) Irp->AssociatedIrp.SystemBuffer;
				
				outputBuffer->Cylinders = devExt->DiskGeometry.Cylinders;
				outputBuffer->MediaType = devExt->DiskGeometry.MediaType;
				outputBuffer->TracksPerCylinder = devExt->DiskGeometry.TracksPerCylinder;
				outputBuffer->SectorsPerTrack = devExt->DiskGeometry.SectorsPerTrack;
				outputBuffer->BytesPerSector = devExt->DiskGeometry.BytesPerSector;
				
				Status = STATUS_SUCCESS;
				Irp->IoStatus.Information = sizeof(DISK_GEOMETRY);		
				break;
			}
		
		case IOCTL_DISK_GET_PARTITION_INFO:
			{
				PPARTITION_INFORMATION outputBuffer;				   
				outputBuffer = ( PPARTITION_INFORMATION )Irp->AssociatedIrp.SystemBuffer;
				
				outputBuffer->PartitionType = PARTITION_FAT_16;
				outputBuffer->BootIndicator = FALSE;
				outputBuffer->RecognizedPartition = TRUE;
				outputBuffer->RewritePartition=FALSE;				
				outputBuffer->StartingOffset = RtlConvertUlongToLargeInteger(0);
				outputBuffer->PartitionLength = RtlConvertUlongToLargeInteger(devExt->DiskLength);
				outputBuffer->HiddenSectors       = (ULONG) (1L);
				outputBuffer->PartitionNumber     = (ULONG) (-1L);
				
				Irp->IoStatus.Status = STATUS_SUCCESS;
				Irp->IoStatus.Information = sizeof(PARTITION_INFORMATION);			
				break;
			}

		case IOCTL_DISK_VERIFY:
			{
				
			PVERIFY_INFORMATION verify_information;

            if (irpSp->Parameters.DeviceIoControl.InputBufferLength <
                sizeof(VERIFY_INFORMATION))
            {
                Status = STATUS_INVALID_PARAMETER;
                Irp->IoStatus.Information = 0;
                break;
            }

            verify_information = (PVERIFY_INFORMATION) Irp->AssociatedIrp.SystemBuffer;

            Status = STATUS_SUCCESS;
            Irp->IoStatus.Information = verify_information->Length;

            break;
			}
		
		case IOCTL_DISK_CHECK_VERIFY:
		case IOCTL_CDROM_CHECK_VERIFY:
		case IOCTL_STORAGE_CHECK_VERIFY:
		case IOCTL_STORAGE_CHECK_VERIFY2:
			{
				Status = STATUS_SUCCESS;
				Irp->IoStatus.Information = 0;
				break;
			}	
		
		case IOCTL_DISK_MEDIA_REMOVAL:
		case IOCTL_STORAGE_MEDIA_REMOVAL:
			{
				Status = STATUS_SUCCESS;
				Irp->IoStatus.Information = 0;
				break;
			}
		
		case IOCTL_DISK_GET_LENGTH_INFO:
			{
				PGET_LENGTH_INFORMATION get_length_information;

				if (irpSp->Parameters.DeviceIoControl.OutputBufferLength < sizeof(GET_LENGTH_INFORMATION)){
					Status = STATUS_BUFFER_TOO_SMALL;
					Irp->IoStatus.Information = 0;
					break;
				}

				get_length_information = (PGET_LENGTH_INFORMATION) Irp->AssociatedIrp.SystemBuffer;				
				get_length_information->Length = RtlConvertUlongToLargeInteger(devExt->DiskLength);

				Status = STATUS_SUCCESS;
				Irp->IoStatus.Information = sizeof(GET_LENGTH_INFORMATION);

				break;
			}
		
		case IOCTL_DISK_IS_WRITABLE:
		{
			Status = STATUS_SUCCESS;
			Irp->IoStatus.Information = 0;
			break;
		}
		
		default:	
			DbgPrint("Unknow Irp ReadWrite %X \n",irpSp->Parameters.DeviceIoControl.IoControlCode);
		break;
		}
		
	Irp->IoStatus.Status=Status;
	IoCompleteRequest(Irp, IO_NO_INCREMENT);

	return Status;
}

NTSTATUS VDevReadWrite (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp){
	PIO_STACK_LOCATION irpSp;
	PUCHAR CurrentAddress;
	PVDEV_EXTENSION   devExt = DeviceObject->DeviceExtension;
	
	irpSp = IoGetCurrentIrpStackLocation(Irp);
	CurrentAddress = MmGetSystemAddressForMdl (Irp->MdlAddress);
	Irp->IoStatus.Information = irpSp->Parameters.Read.Length;
	
	switch (irpSp->MajorFunction)
	{
		case IRP_MJ_READ:			
			RtlMoveMemory(CurrentAddress, devExt->DiskImage + irpSp->Parameters.Read.ByteOffset.LowPart, irpSp->Parameters.Read.Length);
			Irp->IoStatus.Status=STATUS_SUCCESS;
			break;
			
		case IRP_MJ_DEVICE_CONTROL:
			break;
			
		case IRP_MJ_WRITE:
			RtlMoveMemory(devExt->DiskImage + irpSp->Parameters.Read.ByteOffset.LowPart, CurrentAddress,irpSp->Parameters.Read.Length);
			Irp->IoStatus.Status=STATUS_SUCCESS;
			break;
			
		default:
			DbgPrint("Unknow Irp ReadWrite %X \n",irpSp->MajorFunction);
			
			Irp->IoStatus.Status = STATUS_DRIVER_INTERNAL_ERROR;
			break;
	}
	
	
		
	IoCompleteRequest(Irp,(CCHAR) (NT_SUCCESS(Irp->IoStatus.Status) ?IO_DISK_INCREMENT : IO_NO_INCREMENT));
	
	return STATUS_SUCCESS;
}

VOID VDevUnload (IN PDRIVER_OBJECT DriverObject){
	UNICODE_STRING			SymbLink;
	RtlInitUnicodeString(&SymbLink , VSYMB_LINK);
	IoDeleteSymbolicLink(&SymbLink);
	IoDeleteDevice(DriverObject->DeviceObject);
}

NTSTATUS VDevFormatDiskToFat(IN PDEVICE_OBJECT DeviceObject){

    PVDEV_EXTENSION   devExt = DeviceObject->DeviceExtension;
    PBOOT_SECTOR        bootSector = (PBOOT_SECTOR) devExt->DiskImage;
    PUCHAR              firstFatSector;
    ULONG               rootDirEntries;
    ULONG               sectorsPerCluster;
    USHORT              fatType;        // Type FAT 12 or 16
    USHORT              fatEntries;     // Number of cluster entries in FAT
    USHORT              fatSectorCnt;   // Number of sectors for FAT
    PDIR_ENTRY          rootDir;        // Pointer to first entry in root dir
    NTSTATUS            status = STATUS_SUCCESS;

	devExt->DiskRegInfo.RootDirEntries=512;
	devExt->DiskRegInfo.SectorsPerCluster=2;
		
    rootDirEntries = devExt->DiskRegInfo.RootDirEntries;
    sectorsPerCluster = devExt->DiskRegInfo.SectorsPerCluster;

    if (rootDirEntries & (DIR_ENTRIES_PER_SECTOR - 1)) {
        rootDirEntries =
            (rootDirEntries + ( DIR_ENTRIES_PER_SECTOR - 1 )) &
                ~ ( DIR_ENTRIES_PER_SECTOR - 1 );
    }

    bootSector->bsJump[0] = 0xeb;
    bootSector->bsJump[1] = 0x3c;
    bootSector->bsJump[2] = 0x90;
	

    strncpy(bootSector->bsOemName, "RajuRam ", 8);
    bootSector->bsBytesPerSec = (SHORT)devExt->DiskGeometry.BytesPerSector;
    bootSector->bsResSectors  = 1;
    bootSector->bsFATs        = 1;
    bootSector->bsRootDirEnts = (USHORT)rootDirEntries;

    bootSector->bsSectors     = (USHORT)( devExt->DiskRegInfo.DiskSize / devExt->DiskGeometry.BytesPerSector );
    bootSector->bsMedia       = (UCHAR) devExt->DiskGeometry.MediaType;
    bootSector->bsSecPerClus  = (UCHAR)sectorsPerCluster;

    fatEntries =
        (bootSector->bsSectors - bootSector->bsResSectors -
            bootSector->bsRootDirEnts / DIR_ENTRIES_PER_SECTOR) /
                bootSector->bsSecPerClus + 2;

    if (fatEntries > 4087) {
        fatType =  16;
        fatSectorCnt = (fatEntries * 2 + 511) / 512;
        fatEntries -= fatSectorCnt;
        fatSectorCnt = (fatEntries * 2 + 511) / 512;
    }
    else {
        fatType =  12;
        fatSectorCnt = (((fatEntries * 3 + 1) / 2) + 511) / 512;
        fatEntries -= fatSectorCnt;
        fatSectorCnt = (((fatEntries * 3 + 1) / 2) + 511) / 512;
    }

    bootSector->bsFATsecs       = fatSectorCnt;
    bootSector->bsSecPerTrack   = (USHORT)devExt->DiskGeometry.SectorsPerTrack;
    bootSector->bsHeads         = (USHORT)devExt->DiskGeometry.TracksPerCylinder;
    bootSector->bsBootSignature = 0x29;
    bootSector->bsVolumeID      = 0x12345678;
    strncpy(bootSector->bsLabel, "RamDisk    ", 11);
    strncpy(bootSector->bsFileSystemType, "FAT1?   ", 8);
    bootSector->bsFileSystemType[4] = ( fatType == 16 ) ? '6' : '2';
    
    bootSector->bsSig2[0] = 0x55;
    bootSector->bsSig2[1] = 0xAA;

    firstFatSector    = (PUCHAR)(bootSector + 1);
    firstFatSector[0] = (UCHAR) devExt->DiskGeometry.MediaType;
    firstFatSector[1] = 0xFF;
    firstFatSector[2] = 0xFF;

    if (fatType == 16) {
        firstFatSector[3] = 0xFF;
    }

    rootDir = (PDIR_ENTRY)(bootSector + 1 + fatSectorCnt);
    strcpy(rootDir->deName, "MS-RAMDR");
    strcpy(rootDir->deExtension, "IVE");
    rootDir->deAttributes = DIR_ATTR_VOLUME;

    return status;
} 

NTSTATUS FileDiskCreateDevice (IN PDRIVER_OBJECT   DriverObject){
	PVDEV_EXTENSION			devExt;
	PDEVICE_OBJECT			DeviceObject;
	
	UNICODE_STRING			ntuNameString;
	UNICODE_STRING			SymbLink;
	
	NTSTATUS 				Status;
	
	ASSERT(DriverObject != NULL);
	
	RtlInitUnicodeString(&SymbLink , VSYMB_LINK);
	RtlInitUnicodeString(&ntuNameString, VDEV_NAME );
	
	Status = IoCreateSymbolicLink(&SymbLink,&ntuNameString);	
	Status = IoCreateDevice(DriverObject,sizeof(VDEV_EXTENSION),&ntuNameString,FILE_DEVICE_DISK,0,TRUE, &DeviceObject);
	devExt = (PVDEV_EXTENSION) DeviceObject->DeviceExtension;
	
	devExt->DiskLength = (1024*1024);
	devExt->DiskImage = ExAllocatePoolWithTag( NonPagedPool, devExt->DiskLength, VDISK_TAG_DISK);
	DbgPrint("Pointer to allocated mem %X \n",devExt->DiskImage);
	
	ASSERT( devExt->DiskImage != NULL );
	
	devExt->DiskGeometry.BytesPerSector = 512;
    devExt->DiskGeometry.SectorsPerTrack = 32;     
    devExt->DiskGeometry.TracksPerCylinder = 2;
	devExt->DiskGeometry.Cylinders.QuadPart = devExt->DiskLength / 512 / 32 / 2;

	
	devExt->DiskGeometry.MediaType = FixedMedia;			
	
	DeviceObject->Flags |= DO_POWER_PAGABLE;
    DeviceObject->Flags |= DO_DIRECT_IO;	
	
	VDevFormatDiskToFat(DeviceObject);
	
	DeviceObject->Flags &= ~DO_DEVICE_INITIALIZING;	
	
	
	
	Status = STATUS_SUCCESS;	
	return Status;
	
}
